; RS232 Random Number Generator
; Version 2.0
; E. Stuart Hicks
;
; This code for a PIC 12C508 takes entropy data from the amplifier portion of the 
; circuit, debiases it, and sends the resulting byte out to a MAX232 driver chip for 
; TTL -> RS232 voltage scaling.
;
; UART functionality based on free tutorial code from John Massa (www.datadog.com)
; Debiasing based on Bruce Schneier's method from Applied Cryptography 2nd Ed. (p.425)
; Designed in PikLab (piklab.sourceforge.net)
;
; ***************************************************************************
; * This software is protected under the Academic Free License ("AFL") v3.0 *
; * See LICENSE under the root package directory for details                *
; *                                                                         *
; * This code is copyright to E. Stuart Hicks <ehicks@binarymagi.com> and   *
; * the author reserves all rights granted to him under the AFL.            *
; ***************************************************************************

include <p12f508.inc>

; Option bits: TMR0 = counter, prescalar = WDT at 1:128
OPTION_BITS	equ	B'00001111'

; IO pins for GPIO bit tests
INPIN		equ	D'0'	; Entropy comes in GP0 (pin 7)
OUTPIN		equ	D'1'	; RS232 output on GP1 (pin 6)

; Counter value for 9600 baud timing using TIMER0
;TIMECTR        equ     D'83'	; 1 MIPS / (9600 baud / 80% at 8N1) = 83
;TIMECTR        equ     D'104'	; 1 MIPS / 9600 baud = 104
TIMECTR         equ     D'86'	; (1 MIPS / 9600 baud) - (18 cycles between timer test & pin change)

; Constants for the registers used
STOREBUF	equ	H'007'	; buffer for storing data from the input pin (next byte)
STORECTR	equ	H'008'	; shift counter for the # of usable bits currently in the store buffer
SENDBUF		equ	H'009'	; buffer for data currently being sent (last byte)
SENDCTR		equ	H'00A'	; shift counter for the # of bits left to send in the send buffer
FLAGS		equ	H'00B'	; flags register (boolean controls)
CNT		equ	H'00C'	; int counter used in delay loops and such
TIMEREG		equ	H'00D'	; register to hold the TIMECTR value for subtraction
TMP		equ	H'00F'	; register for temporary values

THIS		equ	1	; just so the result operands are a little more readable
F_STORE_FULL	equ	0	; store buffer is full and ready to be transfered to the send buffer
F_SEND_EMPTY	equ	1	; send buffer is empty and ready to be refilled from the store buffer
F_SENT_START	equ	2	; start bit has been sent
F_SENT_STOP	equ	3	; stop bit has been sent

; Configuration: internal clock, watchdog & external clear on, no code protection, code starts at address 0
	__CONFIG _IntRC_OSC & _WDT_ON & _CP_OFF & _MCLRE_ON
	org	0
; Code starts here
start
	movlw	OPTION_BITS		; W <- option settings
	option				; OPTION <- W

	movlw	TIMECTR			; timereg <- TIMECTR
	movwf	TIMEREG
	clrf	STORECTR		; storectr <- 0;
	clrf	SENDCTR			; sendctr <- 0;
	clrf	FLAGS			; <all flags> <- false;
	bsf	FLAGS,F_SEND_EMPTY	; f_send_empty <- true;

	clrf	TMP			; TMP <- 0;
	comf	TMP,THIS		; TMP <- !TMP;
	bcf	TMP,OUTPIN		; TMP[<output pin>] = 0;
	movf	TMP,W			; W <- TMP;
	tris	GPIO			; Set OUTPIN to write mode
	bsf	GPIO,OUTPIN		; <output pin> <- 1; (mark)

; Main loop - branch decision logic & function calls
main_loop				; while(true) {
	clrwdt				;	pet_dog();
	btfss	FLAGS,F_SEND_EMPTY	;	if (!f_sendempty)
	call	send_data		;		send_data();
	btfss	FLAGS,F_STORE_FULL	;	if (!f_storefull)
	call	get_entropy		;		get_entropy();
	btfss	FLAGS,F_STORE_FULL	;	if (f_storefull && f_sendempty)
	goto	main_loop
	btfsc	FLAGS,F_SEND_EMPTY
	call	buffer_swap		;		buffer_swap();
	goto	main_loop		; }

; Get entropy from INPIN
get_entropy
	clrf	TMP			; TMP <- 0;
	btfsc	GPIO,INPIN		; if (<input pin> = 1)
	incf	TMP,THIS		;	++TMP;

	; Delay before next sample
	call	send_data		; send_data();
	call	send_data		; send_data();

	clrw				; W <- 0;
	btfsc	GPIO,INPIN		; if (<input pin> = 1)
	incf	W,W			;	++W;
	xorwf	TMP,THIS		; TMP <- TMP ^ W;
	btfss	TMP,0			; if (TMP = 1)
	return				; 	return;

	movwf	TMP			; TMP <- W
	bcf	STATUS,C		; C <- 0;
	btfsc	TMP,0			; if (LSB(TMP) = 1)
	bsf	STATUS,C		;	C <- 1;
	rlf	STOREBUF,THIS		; storebuf << 1; LSB(storebuf) <- C;
	bsf	STATUS,C		; ++storectr;
	rlf	STORECTR,THIS

	btfsc	STORECTR,7		; if (MSB(storecrt) = 1)
	bsf	FLAGS,F_STORE_FULL 	;	f_storefull <- true;

	return				; return;

; Push store buffer into send buffer & flush to make room for more entropy
buffer_swap
	movf	STOREBUF,W		; sendbuf <- storebuf
	movwf	SENDBUF
	clrf	STOREBUF		; TODO: Remove me after debugging
	clrf	STORECTR		; storectr <- 00000000;
	movlw	B'11111111'		; sendctr <- 11111111;
	movwf	SENDCTR
	bcf	FLAGS,F_STORE_FULL	; f_storefull <- false;
	bcf	FLAGS,F_SEND_EMPTY	; f_sendempty <- false;
	return				; return;

; Sends entropy in SENDBUF
send_data
	; Use the TMR0 register to perform the timings.  It maxs out at 255 (8 bits) but the prescalar
	; can be used to increase the max available time to support slower baud rates.  SUBWF will 
	; allow you to check > values using the carry-over bit in the status register.

	; If it isn't time to send the next bit then jump back to get_entropy
	movf	TMR0,W			; W <- timer0;				;(0)
	subwf	TIMEREG,W		; W <- timereg - W;			;(1)
	btfsc	STATUS,C		; if (W >= 0)				;(2,3)
	return				;	return;

byte_done_test
	; Full reset if we are altogether done with a byte
	btfss	FLAGS,F_SENT_STOP	; if (f_sent_stop)			; (4)
	goto	start_bit_test							; (5,6)
	bcf	FLAGS,F_SENT_START	;	f_sent_start <- false;
	bcf	FLAGS,F_SENT_STOP	;	f_sent_stop <- false;
	bsf	FLAGS,F_SEND_EMPTY	;	f_send_empty <- true;
	return				;	return;

start_bit_test
	; If we haven't started the send yet, set OUTPIN to space & skip over sendbuf
	btfsc	FLAGS,F_SENT_START	; if (!f_sent_start)			; (7)
	goto	stop_bit_test							; (8,9)
	bsf	FLAGS,F_SENT_START	;	f_sent_start <- true;		; (__10)
	nop									; (__11)
	nop									; (__12)
	nop									; (__13)
	nop									; (__14)
	nop									; (__15)
	nop									; (__16)
	nop									; (__17)
	bcf	GPIO,OUTPIN		;	<output pin> <- 0; (space)	; (__18)
	goto	reset_counter		;	goto reset_counter;

stop_bit_test
	; If we're out of bits in sendbuf then send the stop bit
	btfsc	SENDCTR,0		; if (sendctr = 0)			; (10)
	goto	next_bit							; (11,12)
	bsf	FLAGS,F_SENT_STOP	;	f_sent_stop <- true;		; (__13)
	nop									; (__14)
	nop									; (__15)
	nop									; (__16)
	nop									; (__17)
	bsf	GPIO,OUTPIN		;	<output pin> <- 1; (mark)	; (__18)
	goto	reset_counter		;	goto reset_counter;

next_bit
	; If it is then grab the next bit from sendbuf, shift it down, & set OUTPIN accordingly
	bcf	STATUS,C		; --sendctr;				; (13)
	rrf	SENDCTR,THIS							; (14)
	btfss	SENDBUF,0		; if (LSB(sendbuf) = 0)			; (15)
	goto	set_pin_low		;	goto set_pin_low;		; (16,17)
set_pin_high
	nop									; (__17)
	bsf	GPIO,OUTPIN		; <output pin> <- 1; (mark)		; (__18)
	goto	shift_sendbuf		; goto shift_sendbuf;
set_pin_low
	bcf	GPIO,OUTPIN		; <output pin> <- 0; (space)		; (18) = 18 cycles
shift_sendbuf
	bcf	STATUS,C		; TODO: Remove me after debugging
	rrf	SENDBUF,THIS		; sendbuf >> 1;

reset_counter
	; Reset the counter
	clrf	TMR0			; timer0 <- 0;
	return				; return;

END